import argparse
import logging
import sys
import typing
from copy import deepcopy
from datetime import date, datetime

from celus_nigiri.client import Sushi5Client


def parse_date(date_str) -> date:
    return datetime.strptime(date_str, "%Y-%m")


def available_report_types() -> typing.List[str]:
    res = []
    for master_rt in Sushi5Client.report_types:
        res.append(master_rt)
        res.extend(f"{master_rt}_{st}" for st in Sushi5Client.report_types[master_rt]["subreports"])
    return res


def main():
    parser = argparse.ArgumentParser(description="Nigiri downloader")
    parser.add_argument("--debug", "-d", action="store_true", default=False)
    parser.add_argument(
        "--report-type", "-T", metavar="RT", required=True, choices=available_report_types()
    )
    parser.add_argument("--requestor-id", "-R", metavar="ID", required=True)
    parser.add_argument("--customer-id", "-C", metavar="ID", default=None)
    parser.add_argument("--api-key", "-A", metavar="KEY", default=None)
    parser.add_argument("--begin-date", "-B", metavar="YYYY-MM", type=parse_date, required=True)
    parser.add_argument("--end-date", "-E", metavar="YYYY-MM", type=parse_date, required=True)
    parser.add_argument("url", metavar="URL", nargs=1, help="Base URL")
    args = parser.parse_args()

    extra_params = deepcopy(Sushi5Client.EXTRA_PARAMS["maximum_split"].get(args.report_type, {}))
    extra_params.update(Sushi5Client.EXTRA_PARAMS["filters"].get(args.report_type, {}))
    if args.api_key:
        extra_params["api_key"] = args.api_key

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig()

    client = Sushi5Client(
        url=args.url[0],
        requestor_id=args.requestor_id,
        customer_id=args.customer_id,
        extra_params=extra_params,
    )

    response = client.fetch_report_data(args.report_type, args.begin_date, args.end_date)

    if 200 <= response.status_code < 300 or 400 <= response.status_code < 600:
        for data in response.iter_content(1024 * 1024):
            sys.stdout.buffer.write(data)
    else:
        raise RuntimeError(f"Wrong response status code - {response.status_code}")


if __name__ == "__main__":
    main()



import typing
from dataclasses import dataclass, field
from datetime import date


@dataclass
class CounterRecord:
    # The actual vaule
    value: int

    # mandatory, each record should have at least a start date
    start: typing.Optional[date] = None

    # Optional really, if the report is for whole month
    # then the start date (first day of the month) suffice
    end: typing.Optional[date] = None

    # name of the publication
    title: typing.Optional[str] = None

    # ISBN (books), ISSN (periodics), EISSN (same thing only for electronic version of publications)
    title_ids: typing.Dict[str, str] = field(default_factory=dict)

    # name of item
    item: typing.Optional[str] = None

    # ISBN (books), ISSN (periodics), EISSN (same thing only for electronic version of publications)
    item_ids: typing.Dict[str, str] = field(default_factory=dict)

    # contains more details
    dimension_data: typing.Dict[str, str] = field(default_factory=dict)

    # What actually value of this record means
    metric: typing.Optional[str] = None

    # Entity to which are these data related
    organization: typing.Optional[str] = None

    def as_csv(self) -> typing.Tuple[str, ...]:
        def serialize_dict(mapping: typing.Optional[dict]) -> str:
            if not mapping:
                return ""
            return "|".join(f"{k}:{mapping[k]}" for k in sorted(mapping.keys()))

        def format_date(date_obj: typing.Optional[date]):
            if not date_obj:
                return ""
            return date_obj.strftime("%Y-%m-%d")

        return (
            format_date(self.start),
            format_date(self.end),
            self.organization or "",
            self.title or "",
            serialize_dict(self.title_ids),
            self.item or "",
            serialize_dict(self.item_ids),
            serialize_dict(self.dimension_data),
            self.metric or "",
            str(self.value),
        )

    @classmethod
    def empty_generator(cls) -> typing.Generator["CounterRecord", None, None]:
        empty: typing.List["CounterRecord"] = []
        return (e for e in empty)



"""
Module dealing with data in the COUNTER5 format.
"""

import csv
import json
import typing
from datetime import date

import ijson.backends.yajl2_c as ijson

from .celus import get_months as celus_format_get_months
from .csv_detect import detect_csv_dialect, detect_file_encoding
from .error_codes import ErrorCode, error_code_to_severity
from .exceptions import SushiException
from .record import CounterRecord
from .utils import get_date_range, parse_counter_month, parse_date_fuzzy


class CounterError:
    def __init__(self, code=None, severity=None, message=None, data=None):
        self.code = code
        self.severity = severity if isinstance(severity, str) else error_code_to_severity(code)
        self.message = message
        self.data = data

    @classmethod
    def from_sushi_error(cls, sushi_error):
        return cls(
            code=sushi_error.code,
            severity=sushi_error.severity,
            message=sushi_error.text,
            data=sushi_error.data,
        )

    @classmethod
    def from_sushi_dict(cls, rec):
        severity = rec.get("Severity")
        severity = (
            severity if isinstance(severity, str) else error_code_to_severity(rec.get("Code"))
        )
        return cls(
            code=rec.get("Code"),
            severity=rec.get("Severity"),
            message=rec.get("Message"),
            data=rec.get("Data"),
        )

    def to_sushi_dict(self):
        return dict(Code=self.code, Severity=self.severity, Message=self.message, Data=self.data)

    def __str__(self):
        return f'{self.severity or "Error"} #{self.code}: {self.message}'

    def raise_me(self):
        raise SushiException(str(self), content=self.data)


class TransportError:
    def __init__(self, http_status=None, message=None):
        self.http_status = http_status
        self.message = message

    def __str__(self):
        http = f" [HTTP {self.http_status}]" if self.http_status else ""
        return f"Transport error{http}: {self.message}"


ALLOWED_ITEM_IDS = {
    "DR": {"Proprietary": "Proprietary", "Proprietary_ID": "Proprietary"},
    "TR": {
        "DOI": "DOI",
        "Online_ISSN": "Online_ISSN",
        "Print_ISSN": "Print_ISSN",
        "ISBN": "ISBN",
        "Proprietary": "Proprietary",
        "Proprietary_ID": "Proprietary_ID",
        "URI": "URI",
    },
    "PR": {"Proprietary": "Proprietary", "Proprietary_ID": "Proprietary"},
    "IR": {
        "DOI": "DOI",
        "Online_ISSN": "Online_ISSN",
        "Print_ISSN": "Print_ISSN",
        "ISBN": "ISBN",
        "Proprietary": "Proprietary",
        "Proprietary_ID": "Proprietary_ID",
        "URI": "URI",
    },
    "IR_M1": {
        "DOI": "DOI",
        "Online_ISSN": "Online_ISSN",
        "Print_ISSN": "Print_ISSN",
        "ISBN": "ISBN",
        "Proprietary": "Proprietary",
        "Proprietary_ID": "Proprietary_ID",
        "URI": "URI",
    },
}


class Counter5ReportBase:
    dimensions = []  # this should be redefined in subclasses
    allowed_item_ids: typing.Dict[str, str] = {}  # this should be redefined in subclasses

    def __init__(self, report: typing.Optional[typing.IO[bytes]] = None, http_status_code=None):
        self.records = []
        self.queued = False
        self.record_found: bool = False  # is populated once `fd_to_dicts` is called
        self.header = {}
        self.errors: typing.List[typing.Union[CounterError, TransportError]] = []
        self.warnings: typing.List[CounterError] = []
        self.infos: typing.List[CounterError] = []
        self.http_status_code = http_status_code

        # Parse it for the first time to extract errors, warnings and infos
        if report:
            self.fd_to_dicts(report)

    def read_report(
        self, header: dict, items: typing.Generator[dict, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        """
        Reads in the report as returned by the API using Sushi5Client
        """
        for item in items:
            self.check_item(item)

            title = self._item_get_title(item)
            title_ids = self._item_get_title_ids(item)
            item_name = self._item_get_item(item)
            item_ids = self._item_get_item_ids(item)
            dimension_data = self._extract_dimension_data(self.dimensions, item)

            performances = item.get("Performance", [])
            for performance in performances:
                period = performance.get("Period", {})
                start = parse_date_fuzzy(period.get("Begin_Date"))
                end = parse_date_fuzzy(period.get("End_Date"))

                for metric in performance.get("Instance", []):
                    yield CounterRecord(
                        value=int(metric.get("Count")),
                        metric=metric.get("Metric_Type"),
                        start=start,
                        end=end,
                        title=title,
                        title_ids=title_ids,
                        item=item_name,
                        item_ids=item_ids,
                        dimension_data=dimension_data,
                    )

    def check_header(self, header, fd):
        lower_keys = [e.lower() for e in header]
        for field in ["report_id", "customer_id"]:  # mandatory header fields
            if field not in lower_keys:
                raise SushiException("Incorrect format", content=fd.read())

    def check_item(self, item: dict):
        """Check whether the item is valid"""
        self.check_no_extra_ids(item)

    def check_no_extra_ids(self, item: dict):
        """Check that there are no extra title IDs"""
        allowed = {e.lower() for e in self.allowed_item_ids}
        ids_keys = {e["Type"].lower() for e in (item.get("Item_ID", []) or [])}
        extra = ids_keys - allowed
        if extra:
            raise SushiException(f"Extra IDs not allowed {extra}", content=json.dumps(item))

    def extract_errors(self, data):
        from .client import Sushi5Client

        sushi_errors = Sushi5Client.extract_errors_from_data(data)

        if any(str(e.code) == str(ErrorCode.PREPARING_DATA.value) for e in sushi_errors):
            # special status telling us we should retry later
            self.queued = True

        for sushi_error in sushi_errors:
            counter_error = CounterError.from_sushi_error(sushi_error)
            if sushi_error.is_warning:
                self.warnings.append(counter_error)
            elif sushi_error.is_info:
                self.infos.append(counter_error)
            else:
                self.errors.append(counter_error)

    def fd_to_dicts(
        self, fd: typing.IO[bytes]
    ) -> typing.Tuple[dict, typing.Generator[dict, None, None]]:
        def empty_generator() -> typing.Generator[dict, None, None]:
            empty: typing.List[dict] = []
            return (e for e in empty)

        # make sure that fd is at the beginning
        fd.seek(0)

        # first check whether it is not an error report
        # skip potential leading whitespace
        first_character = fd.read(1)  # type: bytes
        while first_character.isspace():
            first_character = fd.read(1)
        fd.seek(0)

        if first_character == b"[":
            # entire json can be a list of errors
            self.extract_errors(json.load(fd))
            return {}, empty_generator()

        elif first_character == b'"':
            # stringified header with an error received
            json_string = json.load(fd)
            data = json.loads(json_string)
            self.extract_errors(data)
            return data, empty_generator()

        elif first_character != b"{":
            self.errors.append(
                TransportError(http_status=self.http_status_code, message="Non SUSHI data returned")
            )
            return {}, empty_generator()

        # Try to read the data
        self.record_found = bool(next(ijson.items(fd, "Report_Items.item"), None))
        fd.seek(0)
        if self.record_found:
            items = ijson.items(fd, "Report_Items.item")
        else:
            self.record_found = bool(next(ijson.items(fd, "body.Report_Items.item"), None))
            if self.record_found:
                items = ijson.items(fd, "body.Report_Items.item")
            else:
                items = empty_generator()
        fd.seek(0)

        # try to read the header
        header = next(ijson.items(fd, "Report_Header"), None)
        fd.seek(0)
        # check whether the header is not located in 'body' element
        if not header:
            header = dict(ijson.kvitems(fd, "body.Report_Header"))
            fd.seek(0)

        if not self.record_found and not header:
            # not data and header not found entire json could be a header
            header = json.load(fd)
        fd.seek(0)

        # Extract errors
        if header:
            # Try to extract exceptions from header
            self.extract_errors(header)
            if not any((self.errors, self.warnings, self.infos)):
                # no  error/warning/info = > file has to contain a valid header
                self.check_header(header, fd)
                fd.seek(0)
        elif not self.record_found:
            # Header is missing and no data
            raise SushiException("Incorrect format", content=fd.read())
        else:
            # Header is empty, but data are present
            pass

        # error can be placed outside of header
        # look for them only if the header is missing or there are no data records
        # - if we look for them in each case, it can add quite some overhead for large files
        #   (like 20s for a 50 MB file)
        if (
            (not header or not self.record_found)
            and not self.errors
            and not self.warnings
            and not self.infos
        ):
            # Extract exceptions from root
            self.extract_errors(list(ijson.items(fd, "Exceptions.item")))  # In <root>
            fd.seek(0)
            self.extract_errors(dict(ijson.kvitems(fd, "Exception")))  # Single exception in <root>
            fd.seek(0)
            # Extract exception(s) from body
            self.extract_errors(list(ijson.items(fd, "body.Exceptions.item")))
            fd.seek(0)
            self.extract_errors(dict(ijson.kvitems(fd, "body.Exception")))
            fd.seek(0)

        return header, items

    def get_months(self, fd: typing.IO[bytes]) -> typing.List[date]:
        pos = fd.tell()

        header, _ = self.fd_to_dicts(fd)

        # Try to obtain End_Date and Begin_Date from report filter
        date_start_str = None
        date_end_str = None
        for e in header.get("Report_Filters", []):
            if e.get("Name") == "Begin_Date":
                date_start_str = e.get("Value")
            elif e.get("Name") == "End_Date":
                date_end_str = e.get("Value")

        fd.seek(pos)

        date_start = parse_date_fuzzy(date_start_str or "")
        date_end = parse_date_fuzzy(date_end_str or "")
        if date_start and date_end:
            return get_date_range(date_start, date_end)
        else:
            return []

    def file_to_records(self, filename: str) -> typing.Generator[CounterRecord, None, None]:
        f = open(filename, "rb")  # file will be closed later (once generator struct is discarded)
        self.header, items = self.fd_to_dicts(f)
        return self.read_report(self.header, items)

    @classmethod
    def file_to_input(cls, filename: str):
        with open(filename, "r", encoding="utf-8") as infile:
            return json.load(infile)

    @classmethod
    def _item_get_title(cls, item):
        return item.get("Title")

    @classmethod
    def _item_get_title_ids(cls, item) -> typing.Dict[str, str]:
        return cls._extract_title_ids(item.get("Item_ID", []) or [])

    @classmethod
    def _item_get_item_ids(self, item) -> typing.Dict[str, str]:
        return {}

    @classmethod
    def _item_get_item(self, item) -> typing.Optional[str]:
        return None

    @classmethod
    def _extract_title_ids(cls, values: list) -> dict:
        ret = {}
        for value in values:
            if id_type := cls.allowed_item_ids.get(value.get("Type")):
                ret[id_type] = value.get("Value")
        return ret

    @classmethod
    def _extract_dimension_data(cls, dimensions: list, record: dict):
        ret = {}
        for dim in dimensions:
            if dim in record:
                ret[dim] = record[dim]
            else:
                # Try to extract dimension from Item_Attributes
                item_attributes = record.get("Item_Attributes", [])
                for item_attribute in item_attributes:
                    if item_attribute.get("Type") == dim:
                        ret[dim] = item_attribute.get("Value", "")
                        break
        return ret


class Counter5DRReport(Counter5ReportBase):
    dimensions = ["Access_Method", "Data_Type", "Publisher", "Platform"]
    allowed_item_ids = ALLOWED_ITEM_IDS["DR"]

    @classmethod
    def _item_get_title(cls, item):
        return item.get("Database")


class Counter5TRReport(Counter5ReportBase):
    dimensions = [
        "Access_Type",
        "Access_Method",
        "Data_Type",
        "Section_Type",
        "YOP",
        "Publisher",
        "Platform",
    ]
    allowed_item_ids = ALLOWED_ITEM_IDS["TR"]


class Counter5PRReport(Counter5ReportBase):
    dimensions = ["Access_Method", "Data_Type", "Platform"]
    allowed_item_ids = ALLOWED_ITEM_IDS["PR"]

    @classmethod
    def _item_get_title(cls, item):
        # there are not titles in the platform report
        return None


class Counter5IRReport(Counter5ReportBase):
    dimensions = [
        "Access_Type",
        "Access_Method",
        "Data_Type",
        "YOP",
        "Publisher",
        "Platform",
        "Article_Version",
        "Parent_Data_Type",
    ]
    allowed_item_ids = ALLOWED_ITEM_IDS["IR"]

    @classmethod
    def _item_get_title(cls, item):
        return item.get("Item_Parent", {}).get("Item_Name", "")

    @classmethod
    def _item_get_item(cls, item):
        return item.get("Item", "")

    @classmethod
    def _item_get_item_ids(cls, item):
        return cls._extract_title_ids(item.get("Item_ID", []) or [])

    @classmethod
    def _item_get_title_ids(cls, item):
        return cls._extract_title_ids(item.get("Item_Parent", {}).get("Item_ID", []) or [])

    @classmethod
    def _extract_dimension_data(cls, dimensions: list, record: dict):
        res = super()._extract_dimension_data(dimensions, record)
        # enrich result with Parent_Data_Type => it can be useful during processing
        if parent_data_type := record.get("Item_Parent", {}).get("Data_Type"):
            res["Parent_Data_Type"] = parent_data_type
        return res


class Counter5IRM1Report(Counter5IRReport):
    dimensions = ["Publisher", "Platform"]
    allowed_item_ids = ALLOWED_ITEM_IDS["IR_M1"]


class Counter5TableReport:
    """
    Implements reading of C5 reports stored in a CSV/TSV tabular format
    """

    dimensions = []

    column_map = {
        "Metric_Type": "metric",
        "Organization": "organization",
        "Database": "title",
        "Title": "title",
        "Item": "title",
    }

    report_type_to_dimensions = {
        "DR": ["Access_Method", "Data_Type", "Publisher", "Platform"],
        "TR": [
            "Access_Type",
            "Access_Method",
            "Data_Type",
            "Section_Type",
            "YOP",
            "Publisher",
            "Platform",
        ],
        "PR": ["Access_Method", "Data_Type", "Platform"],
        "IR": ["Access_Type", "Access_Method", "Data_Type", "YOP", "Publisher", "Platform"],
        "IR_M1": ["Publisher", "Platform"],
    }

    title_id_columns = {
        "DOI": "DOI",
        "ISBN": "ISBN",
        "Print_ISSN": "Print_ISSN",
        "Online_ISSN": "Online_ISSN",
        "Proprietary_ID": "Proprietary",
        "Proprietary": "Proprietary",
        "URI": "URI",
    }

    ignored_columns = ["Reporting_Period_Total", "Publisher_ID", "Linking_ISSN"]

    def file_to_records(self, filename: str) -> typing.Generator[CounterRecord, None, None]:
        # detect encoding
        with open(filename, "rb") as f:
            encoding = detect_file_encoding(f)

        with open(filename, "r", encoding=encoding) as infile:
            # detect dialect
            dialect = detect_csv_dialect(infile)

            for rec in self._fd_to_records(infile, dialect):
                yield rec

    def check_title_ids(self, report_type: str, title_ids: typing.Dict[str, str]):
        extra = set(title_ids.keys()) - set(ALLOWED_ITEM_IDS[report_type])
        if extra:
            raise ValueError(f"Unsupported IDs ({extra}) for report_type {report_type}")

    def get_header(self, reader: csv.reader):
        header = {}
        for header_line in reader:
            if not header_line or not header_line[0].strip():
                # we break on empty line - it means end of header and start of data
                break
            header[header_line[0].strip()] = header_line[1].strip() if len(header_line) > 1 else ""
        return header

    def _fd_to_records(self, infile, dialect) -> typing.Generator[CounterRecord, None, None]:
        # read the header
        reader = csv.reader(infile, dialect=dialect)
        header = self.get_header(reader)

        report_type = header.get("Report_ID")
        if not report_type or report_type not in self.report_type_to_dimensions:
            raise ValueError(f"Unsupported report type: {report_type}")
        if header.get("Release") != "5":
            raise ValueError(f'Unsupported COUNTER release: {header.get("Release")}')

        # we continue reading using a dict reader
        reader = csv.DictReader(infile, dialect=dialect)
        extra_dims = self.report_type_to_dimensions[report_type]
        # process the records
        for record in reader:
            implicit_dimensions = {}
            explicit_dimensions = {}
            monthly_values = {}
            title_ids = {}
            for key, value in record.items():
                key = key.strip()
                value = value.strip()
                if not value:
                    continue
                month = parse_counter_month(key)
                if month:
                    monthly_values[month] = int(value)
                else:
                    if not key or key in self.ignored_columns:
                        pass
                    elif key in self.column_map:
                        implicit_dimensions[self.column_map[key]] = value
                    elif key in extra_dims:
                        explicit_dimensions[key] = value
                    elif key in self.title_id_columns:
                        title_ids[self.title_id_columns[key]] = value
                    else:
                        raise KeyError(f'We don\'t know how to interpret the column "{key}"')

            self.check_title_ids(report_type, title_ids)

            # we put initial data into the data we read
            # - these are usually dimensions that are fixed
            #
            # for the whole import and are not part of the data itself
            # for key, value in initial_data.items():
            #     if key not in implicit_dimensions:
            #         implicit_dimensions[key] = value  # only update if the value is not present
            for month, value in monthly_values.items():
                yield CounterRecord(
                    value=value,
                    start=month,
                    dimension_data=explicit_dimensions,
                    title_ids=title_ids,
                    **implicit_dimensions,
                )

    def get_months(self, fd: typing.IO[bytes], dialect) -> typing.List[date]:
        return celus_format_get_months(fd, self.get_header, dialect)



"""
Module dealing with data in the COUNTER4 format as provided by the pycounter library
"""

import typing

from celus_pycounter import report
from celus_pycounter.report import CounterEresource, CounterReport

from .counter5 import CounterRecord


class Counter4ReportBase:
    dimensions = []  # this should be redefined in subclasses
    allowed_item_ids = ["DOI", "Online_ISSN", "Print_ISSN", "ISBN", "Proprietary"]

    dimension_to_attr = {
        "Publisher": "publisher",
        "Content Provider": "content_provider",
        "Platform": "platform",
    }
    title_id_to_attr = {
        "Print_ISSN": "issn",
        "Online_ISSN": "eissn",
        "ISBN": "isbn",
        "DOI": "doi",
        "Proprietary": "proprietary_id",
    }

    def __init__(self):
        self.records = []
        self.header: dict = {}
        self.record_found: bool = False

    def read_report(self, report: CounterReport) -> typing.Generator[CounterRecord, None, None]:
        """
        Reads in the report as returned by the API using Sushi5Client
        :param report:
        :return:
        """
        self.record_found = len(report.pubs) > 0
        return self._read_report(report)

    def _read_report(self, report: CounterReport) -> typing.Generator[CounterRecord, None, None]:
        for journal in report:  # type: CounterEresource
            title = self._extract_title(journal)
            title_ids = self._extract_title_ids(journal)
            dimension_data = self._extract_dimension_data(self.dimensions, journal)
            for start, metric, value in journal:
                yield CounterRecord(
                    start=start,
                    metric=metric,
                    title=title,
                    title_ids=title_ids,
                    dimension_data=dimension_data,
                    value=value,
                )

    def file_to_records(self, filename: str) -> typing.Generator[CounterRecord, None, None]:
        data = self.file_to_input(filename)

        return self.read_report(data)

    @classmethod
    def file_to_input(cls, filename: str):
        return report.parse(filename)

    def _extract_title(self, record: CounterEresource) -> str:
        return record.collection if hasattr(record, "collection") else record.title

    def _extract_title_ids(self, record) -> dict:
        ret = {}
        for title_id in self.allowed_item_ids:
            attr = self.title_id_to_attr.get(title_id, title_id)
            if hasattr(record, attr):
                ret[title_id] = getattr(record, attr)
        return ret

    @classmethod
    def _extract_dimension_data(cls, dimensions: list, record: CounterEresource) -> dict:
        ret = {}
        for dimension in dimensions:
            attr_name = cls.dimension_to_attr.get(dimension, dimension)
            if hasattr(record, attr_name):
                ret[dimension] = getattr(record, attr_name)
        return ret


class Counter4JR1Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4JR2Report(Counter4ReportBase):
    dimensions = ["Publisher", "Access Denied Category", "Platform"]


class Counter4BR1Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4BR2Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4BR3Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4DB1Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4DB2Report(Counter4ReportBase):
    dimensions = ["Publisher", "Access Denied Category", "Platform"]


class Counter4PR1Report(Counter4ReportBase):
    dimensions = ["Publisher", "Platform"]


class Counter4MR1Report(Counter4ReportBase):
    dimensions = ["Content Provider", "Platform"]



class SushiException(Exception):
    def __init__(self, text, content=None):
        self.text = text
        self.content = content

    def __str__(self):
        return "Sushi exception: {}".format(self.text)



import csv
import json
import logging
import traceback
import typing
import urllib
from io import BytesIO, StringIO
from urllib.parse import urlparse

import requests
import xmltodict
from celus_pycounter import report, sushi
from decouple import Csv, config
from requests import Response

from .counter5 import (
    Counter5DRReport,
    Counter5IRM1Report,
    Counter5IRReport,
    Counter5PRReport,
    Counter5ReportBase,
    Counter5TRReport,
    CounterError,
)
from .error_codes import error_code_to_severity
from .exceptions import SushiException

logger = logging.getLogger(__name__)


class SushiError:
    def __init__(self, code="", text="", full_log="", raw_data=None, severity=None, data=None):
        self.code = code
        self.severity = severity if isinstance(severity, str) else error_code_to_severity(code)
        self.text = text
        self.full_log = full_log
        self.data = data
        self.raw_data = raw_data

    def __str__(self):
        return self.full_log

    @property
    def is_warning(self):
        if self.severity:
            return self.severity.lower() == "warning"

        return False

    @property
    def is_info(self):
        if self.severity:
            return self.severity.lower() == "info"

        return False


def convert_key(key: str) -> str:
    # remove namespace
    key = key.split(":", 1)[-1]

    # to lower
    key = key.lower()

    return key


def recursive_finder(
    data: typing.Any, names: typing.List[str]
) -> typing.Generator[typing.Any, None, None]:
    lower_names = [e.lower() for e in names]

    if isinstance(data, list):
        for e in data:
            for found in recursive_finder(e, names):
                yield found
    elif isinstance(data, dict):
        for key, data in data.items():
            if convert_key(key) in lower_names:
                yield data

            if isinstance(data, (list, dict)):
                for found in recursive_finder(data, names):
                    yield found


class SushiClientBase:
    def __init__(self, url, requestor_id, customer_id=None, extra_params=None, auth=None):
        self.url = url
        self.requestor_id = requestor_id
        self.customer_id = customer_id
        self.extra_params = extra_params
        self.auth = auth

    @classmethod
    def extract_errors_from_data(cls, report_data) -> [SushiError]:
        raise NotImplementedError()

    def report_to_string(self, report_data):
        raise NotImplementedError()

    def get_report_data(
        self,
        report_type,
        begin_date,
        end_date,
        output_content: typing.Optional[typing.IO] = None,
        params=None,
    ):
        raise NotImplementedError()


class Sushi5Client(SushiClientBase):
    """
    Client for SUSHI and COUNTER 5 protocol
    """

    CUSTOMER_ID_PARAM = "customer_id"
    REQUESTOR_ID_PARAM = "requestor_id"

    report_types = {
        "tr": {
            "name": "Title report",
            "subreports": {
                "b1": "Book requests excluding OA_Gold",
                "b2": "Books - access denied",
                "b3": "Book Usage by Access Type",
                "j1": "Journal requests excluding OA_Gold",
                "j2": "Journal articles - access denied",
                "j3": "Journal usage by Access Type",
                "j4": "Journal Requests by YOP (Excluding OA_Gold)",
            },
        },
        "dr": {
            "name": "Database report",
            "subreports": {"d1": "Search and Item usage", "d2": "Database Access Denied"},
        },
        "ir": {
            "name": "Item report",
            "subreports": {"a1": "Journal article requests", "m1": "Multimedia item requests"},
        },
        "pr": {"name": "Platform report", "subreports": {"p1": "View by Metric_Type"}},
    }

    # sets of additional parameters for specific setups
    EXTRA_PARAMS = {
        # split data in TR report to most possible dimensions for most granular data
        "maximum_split": {
            "tr": {"attributes_to_show": "YOP|Access_Method|Access_Type|Data_Type|Section_Type"},
            "ir": {"attributes_to_show": "YOP|Access_Method|Access_Type|Data_Type|Article_Version"},
            "pr": {"attributes_to_show": "Access_Method|Data_Type"},
            "dr": {"attributes_to_show": "Access_Method|Data_Type"},
        },
        "filters": {"ir": {"include_parent_details": True}},
    }

    def __init__(self, url, requestor_id, customer_id=None, extra_params=None, auth=None):
        super().__init__(url, requestor_id, customer_id, extra_params, auth)
        self.session = requests.Session()
        self.session.headers.update(
            {
                "User-Agent": (
                    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:68.0) Gecko/20100101 "
                    "Firefox/68.0"
                )
            }
        )
        proxy = self._get_proxy(url)
        if proxy:
            logger.debug("Using proxy %s for server %s", proxy["proxy"], url)
            proxy_rec = "http://{username}:{password}@{proxy}:{port}/".format(**proxy)
            self.session.proxies.update({"http": proxy_rec, "https": proxy_rec})

    def _get_proxy(self, url):
        parsed = urlparse(url)
        proxy_settings = config(
            "SUSHI_PROXIES", cast=Csv(cast=Csv(post_process=tuple), delimiter=";"), default=""
        )
        for rec in proxy_settings:
            if len(rec) != 5:
                logger.warning("Incorrect proxy definition: [%s]", rec)
                continue
            if not rec[2].isdigit():
                logger.warning("Incorrect proxy port: [%s]", rec[2])
                continue
            if rec[0] == parsed.netloc:
                return {
                    "proxy": rec[1],
                    "port": int(rec[2]),
                    "username": rec[3],
                    "password": rec[4],
                }

    @classmethod
    def _encode_date(cls, value) -> str:
        """
        >>> Sushi5Client._encode_date('2018-02-30')
        '2018-02'
        >>> Sushi5Client._encode_date(datetime(2018, 7, 6, 12, 25, 30))
        '2018-07'
        """
        if hasattr(value, "isoformat"):
            return value.isoformat()[:7]
        return value[:7]

    def _construct_url_params(self, extra=None):
        result = {
            self.CUSTOMER_ID_PARAM: self.customer_id if self.customer_id else self.requestor_id
        }
        if self.requestor_id:
            result[self.REQUESTOR_ID_PARAM] = self.requestor_id
        if self.extra_params:
            result.update(self.extra_params)
        if extra:
            result.update(extra)
        return result

    def _make_request(self, url, params, stream=False):
        logger.debug("Making request to :%s?%s", url, urllib.parse.urlencode(params))
        kwargs = {}
        if self.auth:
            kwargs["auth"] = self.auth
        return self.session.get(url, params=params, stream=stream, **kwargs)

    def get_available_reports_raw(self, params=None) -> bytes:
        """
        Return a list of available reports
        :return:
        """
        url = "/".join([self.url.rstrip("/"), "reports/"])
        params = self._construct_url_params(extra=params)
        response = self._make_request(url, params)
        response.raise_for_status()
        return response.content

    def get_available_reports(self, params=None) -> typing.Generator[dict, None, None]:
        content = self.get_available_reports_raw(params=params)
        reports = self.report_to_data(content)
        return reports

    def make_download_url(self, report_type):
        """
        Prepare download url of a sushi server
        """
        report_type = self._check_report_type(report_type)
        return "/".join([self.url.rstrip("/"), "reports", report_type])

    def make_download_params(self, extra_params, begin_date, end_date):
        """
        Prepare download params which are used to query sushi server
        """
        params = self._construct_url_params(extra=extra_params)
        params["begin_date"] = self._encode_date(begin_date)
        params["end_date"] = self._encode_date(end_date)
        return params

    def fetch_report_data(
        self,
        report_type,
        begin_date,
        end_date,
        dump_file: typing.Optional[typing.IO] = None,
        params=None,
    ) -> Response:
        """
        Makes a request for the data, stores the resulting data into `dump_file` and returns
        the response object for further inspection
        :param report_type:
        :param begin_date:
        :param end_date:
        :param dump_file: where to put file output
        :param params:
        :return:
        """
        url = self.make_download_url(report_type)
        params = self.make_download_params(params, begin_date, end_date)
        response = self._make_request(url, params, stream=bool(dump_file))
        if dump_file is not None:
            for data in response.iter_content(1024 * 1024):
                dump_file.write(data)
            dump_file.seek(0)
        return response

    def get_report_data(
        self,
        report_type,
        begin_date,
        end_date,
        output_content: typing.Optional[typing.IO] = None,
        params=None,
    ) -> Counter5ReportBase:
        response = self.fetch_report_data(
            report_type, begin_date, end_date, params=params, dump_file=output_content
        )
        if 200 <= response.status_code < 300 or 400 <= response.status_code < 600:
            # status codes in the 4xx range may be OK and just provide additional signal
            # about an issue - we need to parse the result in case there is more info
            # in the body
            report_class: typing.Type[Counter5ReportBase]
            report_id = report_type.lower()
            if report_id == "tr":
                report_class = Counter5TRReport
            elif report_id == "dr":
                report_class = Counter5DRReport
            elif report_id == "pr":
                report_class = Counter5PRReport
            elif report_id == "ir":
                report_class = Counter5IRReport
            elif report_id == "ir_m1":
                report_class = Counter5IRM1Report
            else:
                raise NotImplementedError()

            if output_content:
                output_content.seek(0)
            return report_class(output_content, http_status_code=response.status_code)
        # for other response codes we raise an error - it should be only exotic ones
        response.raise_for_status()

    def report_to_data(self, report: bytes, validate=True) -> typing.Generator[dict, None, None]:
        try:
            fd = BytesIO(report)
            counter_report = Counter5ReportBase()
            header, data = counter_report.fd_to_dicts(fd)
        except ValueError as e:
            raise SushiException(str(e), content=report)
        if validate:
            self.validate_data(counter_report.errors, counter_report.warnings)
        return data

    @classmethod
    def validate_data(cls, errors: typing.List[CounterError], warnings: typing.List[CounterError]):
        """Checks that the parsed erorrs and warings are not fatal.
            If so, it will raise SushiException
        :param errors: list of errors
        :param warnings: list of warnings
        :return:
        """
        if len(errors) == 1:
            errors[0].raise_me()
        elif len(errors) >= 1:
            message = "; ".join(
                cls._format_error(error.to_sushi_dict()).full_log for error in errors
            )
            raise SushiException(message, content=[e.data for e in errors])

        # log warnings
        for warning in warnings:
            logging.warning(
                "Warning Exception in report: %s", cls._format_error(warning.to_sushi_dict())
            )

    @classmethod
    def extract_errors_from_data(cls, report_data: typing.Union[dict, list]):
        errors: typing.List[SushiError] = []

        # extract exceptions from list
        if isinstance(report_data, list):
            errors = []
            for item in report_data:
                errors += cls.extract_errors_from_data(item)
            return errors

        # exception in Exception object
        for exception in recursive_finder(report_data, ["Exception"]):
            errors.append(cls._format_error(exception))

        # exceptions in Exceptions object
        for exceptions in recursive_finder(report_data, ["Exceptions"]):
            for exception in exceptions:
                errors.append(cls._format_error(exception))

        if not errors:
            # naked exception in root
            naked = cls._format_error(report_data)

            # at least severity and code needs to be present
            if naked.code and naked.severity:
                errors.append(naked)

        return errors

    @classmethod
    def _format_error(cls, exc: dict):
        code = next(recursive_finder(exc, ["Code"]), None)
        severity = error_code_to_severity(code)
        text = next(recursive_finder(exc, ["Message"]), "")
        data = next(recursive_finder(exc, ["Data"]), "")

        if code is None:
            # Some responses contains "number" instead of "code"
            code = next(recursive_finder(exc, ["number"]), "")

        message = f"{severity} error {code}: {text}"
        if data:
            message += f"; {data}"

        error = SushiError(
            code=code, text=text, full_log=message, severity=severity, raw_data=exc, data=data
        )
        return error

    def _check_report_type(self, report_type):
        report_type = report_type.lower()
        if "_" in report_type:
            main_type, subtype = report_type.split("_", 1)
        else:
            main_type = report_type
            subtype = None
        if main_type not in self.report_types:
            raise ValueError(f"Report type {main_type} is not supported.")
        if subtype and subtype not in self.report_types[main_type]["subreports"]:
            raise ValueError(f"Report subtype {subtype} is not supported for type {main_type}.")
        return report_type

    def report_to_string(self, report_data):
        return json.dumps(report_data, ensure_ascii=False, indent=2)


class Sushi4Client(SushiClientBase):
    """
    Client for SUSHI and COUNTER 4 protocol - a simple proxy for the celus_pycounter.sushi
    implementation
    """

    @staticmethod
    def to_pycounter_report_type(celus_report_type: str) -> str:
        """Celus and pycounter report types may differ"""
        CONVERSIONS = {"JR1GOA": "JR1GOA"}
        return CONVERSIONS.get(celus_report_type, celus_report_type)

    # remap of extra params into names that have special meaning for the pycounter client
    extra_params_remap = {"Name": "requestor_name", "Email": "requestor_email"}

    def __init__(self, url, requestor_id, customer_id=None, extra_params=None, auth=None):
        super().__init__(url, requestor_id, customer_id, extra_params, auth)

    @classmethod
    def _encode_date(cls, value) -> str:
        if hasattr(value, "isoformat"):
            return value.isoformat()[:7]
        return value[:7]

    def get_report_data(
        self,
        report_type,
        begin_date,
        end_date,
        output_content: typing.Optional[typing.IO] = None,
        params=None,
    ) -> report.CounterReport:
        kwargs = {"customer_reference": self.customer_id}
        if self.requestor_id:
            kwargs["requestor_id"] = self.requestor_id
        if params:
            # recode params that have special meaning
            for orig, new in self.extra_params_remap.items():
                if orig in params:
                    kwargs[new] = params.pop(orig)
            # put the rest in as it is
            kwargs.update(params)
        if self.auth:
            kwargs["auth"] = self.auth

        report_type = Sushi4Client.to_pycounter_report_type(report_type)
        report = sushi.get_report(
            self.url, begin_date, end_date, report=report_type, dump_file=output_content, **kwargs
        )
        return report

    @classmethod
    def extract_errors_from_data(cls, report_data: typing.IO[bytes]):
        try:
            report_data.seek(0)  # set to start
            content = report_data.read().decode("utf8", "ignore")
            parsed = xmltodict.parse(content)
        except Exception as e:
            log = f"Exception: {e}\nTraceback: {traceback.format_exc()}"
            return [SushiError(code="non-sushi", text=str(e), full_log=log, severity="Exception")]
        else:
            errors: typing.List[SushiError] = []
            for exception in recursive_finder(parsed, ["Exception"]):
                if not isinstance(exception, dict):
                    # skip non-object exceptions
                    continue
                code = next(recursive_finder(exception, ["Number"]), "")
                code = code.get("#text", "") if isinstance(code, dict) else code
                message = next(recursive_finder(exception, ["Message"]), "")
                message = message.get("#text", "") if isinstance(message, dict) else message
                severity = error_code_to_severity(code)

                full_log = f"{severity}: #{code}; {message}"
                errors.append(
                    SushiError(
                        code=code,
                        text=message,
                        severity=severity,
                        full_log=full_log,
                        raw_data=exception,
                    )
                )

            if not errors:
                errors.append(
                    SushiError(
                        code="non-sushi",
                        text="Could not find Exception data in XML, probably wrong format",
                        full_log="Could not find Exception data in XML, probably wrong format",
                        severity="Exception",
                    )
                )
            return errors

    def report_to_string(self, report_data: report.CounterReport):
        lines = report_data.as_generic()
        out = StringIO()
        writer = csv.writer(out, dialect="excel", delimiter="\t")
        writer.writerows(lines)
        return out.getvalue()



import itertools
import pathlib
import sys

from . import csv_line_length_histogram, detect_csv_dialect, detect_file_encoding


def main():
    for path in itertools.chain(
        pathlib.Path(sys.argv[1]).glob("**/*.csv"), pathlib.Path(sys.argv[1]).glob("**/*.tsv")
    ):
        with path.open("rb") as f:
            encoding = detect_file_encoding(f)

        try:
            with path.open("r", encoding=encoding) as f:
                dialect = detect_csv_dialect(f)
                histogram = csv_line_length_histogram(f, dialect)
        except UnicodeDecodeError as e:
            print(f"{path} {e}")
        else:
            print(f"{path}: {dialect}|{encoding} - {histogram}")


if __name__ == "__main__":
    main()
